# Implémentation

<!--

## TODO

- Ajouter suppression d'un fichier
  - méthode rapide : seul le bitmap des inode et des blocs est modifié
  - méthode sécurisée : les blocs de données et le contenus des inodes sont également effacés

-->

--

# Adressage des blocs de données


--

## Facilité d'adressage

On désire obtenir un accès ordonné aux blocs qui composent le contenu d'un fichier : début du fichier en bloc 0, suite en bloc 1, etc.

![](images/blocks_simple_addr.png) <!-- .element width="60%" -->

--

## Adressage : problème 1

Les blocs de données d'un fichier ne sont pas toujours contigus sur le disque

![](images/blocks_addr_problem.png) <!-- .element width="80%" -->


--

## Adressage souhaité

On désire une fonction `bmap` qui mappe un numéro de bloc du fichier (0, 1, 2, ...) vers un numéro de bloc du disque :

```c
// i : numero du bloc (logique) du fichier
// renvoie le numero du bloc physique sur disque
int bmap(inode, i);
```

![](images/bmap.png)<!-- .element width="50%" -->


<small>

- `bmap(inode, 0) $\rightarrow$ 18`
- `bmap(inode, 1) $\rightarrow$ 20`
- `bmap(inode, 2) $\rightarrow$ 25`
- `bmap(inode, 3) $\rightarrow$ 11`

</small>

--

## Adressage : problème 2

Certains blocs peuvent être des blocs indirects, c'est à dire contenir les adresses d'autres blocs

![](images/blocks_addr_problem2.png) <!-- .element width="80%" -->

--

## Problème des indirections

- Selon le même principe, on peut ajouter des blocs doublement, triplement, etc. indirects
- La distinction entre bloc simple, bloc indirect et bloc double indirect se fait sur sa position dans le tableau des adresses de l'inode

![](images/blocks_indirections.png)<!-- .element width="70%" -->

--

## Repenser la structure de données

Ainsi la structure des inodes de MINIX :

```c
struct minix_inode {
    ...

    u16 i_zone[7];         // direct pointers
    u16 i_indir_zone;      // indirect pointer
    u16 i_dbl_indir_zone;  // double indirect pointer
};
```

peut s'écrire :

```c
struct minix_inode {
    ...

    u16 i_zone[9];  // i_zone[7] = indirect pointers
                    // i_zone[8] = double indirect pointers
};
```

--

## Fonction bmap : algorithme (1/2)

```c
// i : numero du bloc (logique) du fichier
// renvoie le numero du bloc physique sur disque
int bmap(inode, i);
```

Pseudo-code avec 2 niveaux d'indirections :

```c
if (i < nb_blocs_adressables_directement) {  // cas 1
    return inode.i_zone[i]
}

i = i-nb_blocs_adressables_directement

if (i < nb_blocs_adressables_avec_une_seule_indir) {  // cas 2
    indir_bloc = read_block(inode.i_zone[indice_du_bloc_des_indir_simples])
    return indir_bloc[i]
}

i = i-nb_blocs_adressables_avec_une_seule_indir

if (i < nb_blocs_adressables_avec_une_double_indir) {  // cas 3
    indir_bloc = read_block(inode.i_zone[indice_du_bloc_des_indir_doubles])
    indir2_bloc = read_block(
                      indir_bloc[i/nb_blocs_adressables_avec_une_seule_indir])
    return indir2_bloc[i % nb_blocs_adressables_avec_une_seule_indir]
}
```
<!-- .element style="font-size:0.45em" -->

--

## Fonction bmap : algorithme (2/2)

- Pourquoi la soustraction entre chaque cas ? \
pour réajuster l'indice dans le bloc lors d'une indirection : on passe d'un indice absolu à un indice relatif au bloc indirect

<br></br>

- Exemple : si 7 blocs sont adressables directement, mais qu'on veut le 8ème, on veut le premier indice du bloc
indirect, pas l'indice 8

--

## Fonction bmap : exemple (1/4)

- Nombre de blocs adressables directement : 2
- Nombre de blocs adressables avec une seule indirection : 4
- Nombre de blocs adressables avec une double indirection : 4*4

![](images/bmap_example.png)<!-- .element width="65%" -->

<small> ⚠️ ici les blocs sont indéxés logiquement en réalité c'est leur numéro physique qui est présents dans les tableaux de zone

--

## Fonction bmap : exemple (2/4)

```c
if (i < 2) {  // cas 1
  return inode.i_zone[i]
}

i = i-2

if (i < 4) {  // cas 2
  indir_bloc = read_block(inode.i_zone[indice_du_bloc_des_indir_simples])
  return indir_bloc[i]
}

i = i-4

if (i < 16) {  // cas 3
  indir_bloc = read_block(inode.i_zone[indice_du_bloc_des_indir_doubles])
  indir2_bloc = read_block(indir_bloc[i/4])
  return indir2_bloc[i % 4]
}
```
<!-- .element style="font-size:0.45em" -->

--

## Fonction bmap : exemple (3/4)

<div class="text-block" style="font-size: 0.7em">

**i = 4** :

- pas à inférieur à *2* $\rightarrow$ on rejette le cas 1
- 4-*2* = **2**
- **2** inférieur à *4* $\rightarrow$ on entre dans le cas 2
- lecture du bloc d'indirection simple
- indice **2** dans bloc d'indirection simple pointe bien vers bloc 4

**i = 6** :

- pas inférieur à *2* $\rightarrow$ on rejette le cas 1
- 6-*2* = **4**
- **4** pas inférieur à **4** $\rightarrow$ on rejette le cas 2
- **4**-**4** = <text style="color:red">**0**</text>
- **<text style="color:red">0</text>** inférieur à *16* $\rightarrow$ on entre dans le cas 3
- lecture du bloc d'indirection simple d'indice **<text style="color:red">0</text>**/*4* = 0
  - **<text style="color:red">0</text>** % *4* (0) dans bloc d'indirection simple pointe bien vers bloc 6

</div>

--

## Fonction bmap : exemple (4/4)

<div class="text-block" style="font-size: 0.7em">

**i = 9** :

- pas inférieur à *2* $\rightarrow$ on rejette le cas 1
- 9-*2* = **7**
- **7** pas inférieur à *4* $\rightarrow$ on rejette le cas 2
- **7**-4 = **<text style="color:red">3</text>**
- **<text style="color:red">3</text>** inférieur à *16* $\rightarrow$ on entre dans le cas 3
- lecture du bloc d'indirection simple d'indice **<text style="color:red">3</text>**/*4* = 0
  - \scriptsize indice **<text style="color:red">3</text>** % *4* (3) dans bloc d'indirection simple pointe bien vers bloc 9

**i = 10** :

- pas inférieur à *2* $\rightarrow$ on rejette le cas 1
- 10-*2* = **<text style="color:red">8</text>**
- **<text style="color:red">8</text>** pas inférieur à *4* $\rightarrow$ on rejette le cas 2
- **<text style="color:red">8</text>**-*4* = **<text style="color:red">4</text>**
- **<text style="color:red">4</text>** inférieur à *16* $\rightarrow$ on entre dans le cas 3
- lecture du bloc d'indirection simple d'indice **<text style="color:red">4</text>**/*4* = 1
  - \scriptsize indice **<text style="color:red">4</text>** % *4* (0) dans bloc d'indirection simple pointe bien vers bloc 10

</div>

--

# Résolution du chemin d'accès

--

## Problème du chemin d'accès

<br></br>

- Les inodes n'ont pas de noms, juste un numéro...

<br></br>

- Comment retrouver le numéro d'un inode à partir de son chemin d'accès dans la chaîne des répertoires du système de fichiers ?

--

## Fonction namei

On désire une fonction `namei` qui retourne le numéro d'un inode à partir de son chemin d'accès dans le système de fichiers :

```c
// renvoie le numero d'inode du fichier dont le
// chemin d'acces est path
int namei(path);
```

- La recherche débute depuis :
  - l'inode de la racine si le chemin d'accès est absolu
  - le répertoire courant du processus si le chemin d'accès est relatif
- `namei` est typiquement utilisée par tous les appels systèmes qui prennent un chemin d'accès en argument

--

## Fonction namei : example

P.ex. appeler `namei` sur le chemin `"/usr/bin/bash"` parcourt le contenu des inodes suivants et renvoie 21 :

![](images/namei_example.png)<!-- .element width="80%" -->

--

## Fonction namei : algorithme

Pseudo-code pour un chemin absolu :

```c
int namei(path) {
    inode = ROOT_INODE
    pour chaque composant du path p {
        inode = lookup_entry(inode, p)
    }
    return inode
}
```

--

## Fonction lookup_entry

```c
// renvoie le numero d'inode correspondant a name
int lookup_entry(dir_inode, name);
```

- La fonction `lookup_entry` recherche un fichier dans un répertoire et renvoie son inode
- `lookup_entry` réalise une recherche linéaire de la chaîne de caractères `name` dans les entrées de répertoire (`dir_entry`) du répertoire dont l'inode est `dir_inode` :

<fieldset class="warning">
<legend>Attention:</legend>

- Le contenu du répertoire peut-être réparti sur plusieurs blocs `$\rightarrow$` utiliser `bmap`
- Aucune garantie que la taille du répertoire soit un multiple de la taille des blocs : utiliser le champs taille de l'inode pour limiter la recherche

</fieldset>

--

## Fonction lookup_entry : algorithme


```c
int lookup_entry(dir_inode, name) {
    pour chaque bloc de dir_inode et dans la limite
    de dir_inode.size lire le prochain dir_entry {
        if (dir_entry.name == name) {
            return dir_entry.inode
        }
    }
    return not_found
}
```

--

## Fonction del_entry

```c
del_entry(dir_inode, name)
```
<div style="font-size: 0.8em">

- La fonction `del_entry` supprime un fichier d'un répertoire
- `del_entry` réalise une recherche linéaire de la chaîne de caractères `name` dans les entrées de répertoire (`dir_entry`) du répertoire dont l'inode est `dir_inode` :
- Attention : tout comme pour `lookup_entry`, le contenu du répertoire peut-être réparti sur plusieurs blocs
- Une fois l'entrée de répertoire (`dir_entry`) et son numéro d'inode trouvée:
  - le nombre de liens pointant vers l'inode correspondant est décrémenté;
  - si le nombre de liens tombe à 0 il faut supprimer l'inode (e.g. en mettant le bit dans le bitmap d'inode à 0);
  - le champ inode du `dir_entry` est mis à zéro et le bloc modifié est écrit sur disque.

</div>